import numpy as np
from ast import literal_eval
from spacy.gold import biluo_tags_from_offsets, tags_to_entities
from wasabi import Printer
from scipy.sparse import coo_matrix
messenger = Printer()


def get_default_columns():
    """
    To get list of columns names to be default in annotation dataframe.
    :return: list of columns names.
    """
    return [
        'tokens.id',
        'tokens.text',
        'tokens.start',
        'tokens.end',
        'tokens.pos',
        'tokens.tag',
        'tokens.dep',
        'tokens.head',
        'ents.start',
        'ents.end',
        'ents.label']


def map_user_columns_names_with_default(user_columns):
    """
    To map user provided column names to specific format
    :param user_columns: list, of user provided columns
    :return: list, of specifically formatted column names
    """
    column_map = {
        "id": "tokens.id",
        "start": "tokens.start",
        "end": "tokens.end",
        "pos": "tokens.pos",
        "tag": "tokens.tag",
        "dep": "tokens.tag",
        "head": "tokens.head",
        "text": "tokens.text",
        "lemma": "tokens.lemma",
        "lower": "tokens.lower",
        "shape": "tokens.shape",
        "prefix": "tokens.prefix",
        "suffix": "tokens.suffix",
        "is_alpha": "tokens.is_alpha",
        "is_ascii": "tokens.is_ascii",
        "is_digit": "tokens.is_digit",
        "is_lower": "tokens.is_lower",
        "is_upper": "tokens.is_upper",
        "is_title": "tokens.is_title",
        "is_punct": "tokens.is_punct",
        "is_left_punct": "tokens.is_left_punct",
        "is_right_punct": "tokens.is_right_punct",
        "is_space": "tokens.is_space",
        "is_bracket": "tokens.is_bracket",
        "is_quote": "tokens.is_quote",
        "is_currency": "tokens.is_currency",
        "like_url": "tokens.like_url",
        "like_num": "tokens.like_num",
        "like_email": "tokens.like_email",
        "is_oov": "tokens.is_oov",
        "is_stop": "tokens.is_stop",
        "ancestors": "tokens.ancestors",
        "conjuncts": "tokens.conjuncts",
        "children": "tokens.children",
        "lefts": "tokens.lefts",
        "rights": "tokens.rights",
        "n_lefts": "tokens.n_lefts",
        "n_rights": "tokens.n_rights",
        "is_sent_start": "tokens.is_sent_start",
        "has_vector": "tokens.has_vector",
        "vector": "tokens.vector",
        "ent_start": "ents.start",
        "ent_end": "ents.end",
        "ent_label": "ents.label"}

    user_defined_columns = []

    for uc in user_columns:
        try:
            user_defined_columns.append(column_map[uc])
        except KeyError:
            messenger.fail("could not recognize given column name:'{}' so skipping it".format(uc))
            continue

    default_columns = get_default_columns()

    additional_attributes = any([True if nc not in default_columns else False for nc in user_defined_columns])

    return user_defined_columns, additional_attributes


def additional_attributes_map(column):
    """
    To map column names of additional attributes to match spaCy's Token API.
    :param column: list, of column names
    :return: tuple, containing API info
    """
    attributes_map = {
        "tokens.lemma": ("lemma_", False, 0),
        "tokens.lower": ("lower_", False, 0),
        "tokens.shape": ("shape_", False, 0),
        "tokens.prefix": ("prefix_", False, 0),
        "tokens.suffix": ("suffix_", False, 0),
        "tokens.is_alpha": ("is_alpha", False, 0),
        "tokens.is_ascii": ("is_ascii", False, 0),
        "tokens.is_digit": ("is_digit", False, 0),
        "tokens.is_lower": ("is_lower", False, 0),
        "tokens.is_upper": ("is_upper", False, 0),
        "tokens.is_title": ("is_title", False, 0),
        "tokens.is_punct": ("is_punct", False, 0),
        "tokens.is_left_punct": ("is_left_punct", False, 0),
        "tokens.is_right_punct": ("is_right_punct", False, 0),
        "tokens.is_space": ("is_space", False, 0),
        "tokens.is_bracket": ("is_bracket", False, 0),
        "tokens.is_quote": ("is_quote", False, 0),
        "tokens.is_currency": ("is_currency", False, 0),
        "tokens.like_url": ("like_url", False, 0),
        "tokens.like_num": ("like_num", False, 0),
        "tokens.like_email": ("like_email", False, 0),
        "tokens.is_oov": ("is_oov", False, 0),
        "tokens.is_stop": ("is_stop", False, 0),
        "tokens.ancestors": ("ancestors", True, 2),
        "tokens.conjuncts": ("conjuncts", True, 2),
        "tokens.children": ("children", True, 2),
        "tokens.lefts": ("lefts", True, 2),
        "tokens.rights": ("rights", True, 2),
        "tokens.n_lefts": ("n_lefts", False, 2),
        "tokens.n_rights": ("n_rights", False, 2),
        "tokens.is_sent_start": ("is_sent_start", False, 2),
        "tokens.has_vector": ("has_vector", False, 2),
        "tokens.vector": ("vector", False, 2)
    }

    return attributes_map[column]


def merge_entity_details(json_doc):
    """
    To merge entity and token details provided by Doc.to_json() method.
    :param json_doc: dictionary, returned by Doc.to_json()
    :return: new dictionary
    """
    if "ents" in json_doc and json_doc["ents"]:
        ents_dict = {str(ent["start"]) + "_" + str(ent["end"]): ent for ent in json_doc["ents"]}

        new_tokens_data_list = []

        for token_data in json_doc["tokens"]:
            if str(token_data["start"]) + "_" + str(token_data["end"]) in ents_dict:
                token_data["label"] = ents_dict[str(token_data["start"]) + "_" + str(token_data["end"])]["label"]
                new_tokens_data_list.append(token_data)
            else:
                token_data["label"] = None
                new_tokens_data_list.append(token_data)
        json_doc["tokens"] = new_tokens_data_list
        json_doc["ents"] = True
        return json_doc
    else:
        json_doc["ents"] = False
        return json_doc


def get_training_pipeline_from_column_names(columns):
    """
    To infer training pipeline from column names provided by training CSV/XLS.
    :param columns: list, of input dataframe column names
    :return: str, training pipeline
    """
    columns = set(columns)
    _all = {"text", "token_orth", "token_tag", "token_head", "token_dep", "entities"}
    _only_tagger = {"text", "token_orth"}
    _only_parser = {"text", "token_head"}
    _only_ner = {"text", "entities"}
    _tagger_and_parser = {"text", "token_orth", "token_tag", "token_head", "token_dep"}

    if len(columns & _all) == len(_all):
        return "tagger,parser,ner"
    elif len(columns & _only_tagger) == len(_only_tagger):
        return "tagger"
    elif len(columns & _only_parser) == len(_only_tagger):
        return "parser"
    elif len(columns & _only_ner) == len(_only_ner):
        return "ner"
    else:
        return None


def entity_offset_to_biluo_format(nlp, rows, pipline):
    """
    To convert entity offset (start, end) into BILUO format.
    :param nlp: nlp pipeline
    :param rows: dataframe rows (text)
    :param pipline: list, training pipline inferred from data
    :return: list of tuples, containing annotated text and entity tag info
    """
    biluo_rows = []
    default_pipline = ["tagger", "ner"]
    disabled_components = tuple(set(default_pipline) & set(pipline))
    with nlp.disable_pipes(*disabled_components):
        for row in rows.iterrows():
            doc = nlp(row[1]["text"])
            if "ner" in pipline:
                entities = literal_eval(row[1]["entities"])
                tags = biluo_tags_from_offsets(doc, entities)
                if entities:
                    for start, end, label in entities:
                        span = doc.char_span(start, end, label=label)
                        if span and span not in doc.ents:
                            doc.ents = list(doc.ents) + [span]
                if doc.ents:
                    biluo_rows.append((doc, tags))
            else:
                biluo_rows.append((doc, None))
    return biluo_rows


def dataframe_to_spacy_training_json_format(dataframe, nlp, pipline):
    """
    To convert dataframe into spaCy's CLI training JSON format.
    :param dataframe: dataframe, training/validation data
    :param nlp: nlp pipeline
    :param pipline: str, training pipeline
    :return: JSON object, containing training data
    """
    pipline = pipline.split(",")
    list_of_documents = []
    biluo_rows = entity_offset_to_biluo_format(nlp, dataframe, pipline)

    for _id, biluo_row in enumerate(biluo_rows):
        doc, tags = biluo_row
        doc_sentences = []
        if "tagger" in pipline and "parser" in pipline and "ner" in pipline:
            document_row = dataframe.iloc[[_id]]
            token_orth = document_row["token_orth"].iloc[0].replace("'", "").split(", ")
            token_tag = document_row["token_tag"].iloc[0].replace("'", "").split(", ")
            token_head = document_row["token_head"].iloc[0].replace("'", "").split(", ")
            token_dep = document_row["token_dep"].iloc[0].replace("'", "").split(", ")
            tags_to_entities(tags)
            for sentence in doc.sents:
                sentence_tokens = []

                assert len(sentence) <= len(token_orth), messenger.fail("number of token and token_orth field mismatch")
                assert len(sentence) <= len(token_tag), messenger.fail("number of token and token_tag field mismatch")
                assert len(sentence) <= len(token_head), messenger.fail("number of token and token_head field mismatch")
                assert len(sentence) <= len(token_dep), messenger.fail("number of token and token_dep field mismatch")

                for token_id, token in enumerate(sentence):
                    token_data = {
                        "id": token_id,
                        "orth": token_orth[token_id],
                        "tag": token_tag[token_id],
                        "head": int(token_head[token_id]),
                        "dep": token_dep[token_id],
                        "ner": tags[token_id]
                    }
                    sentence_tokens.append(token_data)
                doc_sentences.append({"tokens": sentence_tokens})
        elif "tagger" in pipline:
            document_row = dataframe.iloc[[_id]]
            token_orth = document_row["token_orth"].iloc[0].replace("'", "").split(", ")
            token_tag = document_row["token_tag"].iloc[0].replace("'", "").split(", ")
            for sentence in doc.sents:
                sentence_tokens = []

                assert len(sentence) <= len(token_orth), messenger.fail("number of token and token_orth field mismatch")
                assert len(sentence) <= len(token_tag), messenger.fail("number of token and token_tag field mismatch")

                for token_id, token in enumerate(sentence):
                    token_data = {
                        "id": token_id,
                        "orth": token_orth[token_id],
                        "tag": token_tag[token_id]
                    }
                    sentence_tokens.append(token_data)
                doc_sentences.append({"tokens": sentence_tokens})
        elif "parser" in pipline:
            document_row = dataframe.iloc[[_id]]
            token_head = document_row["token_head"].iloc[0].replace("'", "").split(", ")
            token_dep = document_row["token_dep"].iloc[0].replace("'", "").split(", ")
            for sentence in doc.sents:
                sentence_tokens = []

                assert len(sentence) <= len(token_head), messenger.fail("number of token and token_head field mismatch")
                assert len(sentence) <= len(token_dep), messenger.fail("number of token and token_dep field mismatch")

                for token_id, token in enumerate(sentence):
                    token_data = {
                        "id": token.i,
                        "orth": token.orth_,
                        "tag": token.tag_,
                        "head": int(token_head[token_id]),
                        "dep": token_dep[token_id]
                    }
                    sentence_tokens.append(token_data)
                doc_sentences.append({"tokens": sentence_tokens})
        elif "ner" in pipline:
            for sentence in doc.sents:
                sentence_tokens = []
                for token in sentence:
                    token_data = {
                        "id": token.i,
                        "orth": token.orth_,
                        "tag": token.tag_,
                        "head": token.head.i - token.i,
                        "dep": token.dep_,
                        "ner": tags[token.i]
                    }
                    sentence_tokens.append(token_data)
                doc_sentences.append({"tokens": sentence_tokens})

        list_of_documents.append({
            "id": _id,
            "paragraphs": [
                {
                    "raw": doc.text,
                    "sentences": doc_sentences
                }
            ]
        })
    return list_of_documents


def confusion_matrix(prediction, target, label_map):
    inverse_label_map = {v: k for k, v in label_map.items()}
    prediction = np.asarray([inverse_label_map[i] for i in prediction])
    target = np.asarray([inverse_label_map[i] for i in target])

    if len(label_map) == 2:
        return coo_matrix((np.ones(target.shape[0], dtype=np.int64), (target, prediction)),
                          shape=(len(label_map), len(label_map))).toarray()

    true_positives = prediction == target
    true_positives_bins = target[true_positives]

    if len(true_positives_bins):
        tp_sum = np.bincount(true_positives_bins, minlength=len(label_map))
    else:
        true_sum = prediction_sum = tp_sum = np.zeros(len(label_map))

    if len(prediction):
        prediction_sum = np.bincount(prediction, minlength=len(label_map))
    else:
        prediction_sum = np.zeros(len(label_map))

    if len(target):
        true_sum = np.bincount(target, minlength=len(label_map))
    else:
        true_sum = np.zeros(len(label_map))

    fp = prediction_sum - tp_sum
    fn = true_sum - tp_sum
    tp = tp_sum
    tn = target.shape[0] - tp - fp - fn

    return np.array([tn, fp, fn, tp]).T.reshape(-1, 2, 2)


def precision_recall_fscore(_confusion_matrix):

    if _confusion_matrix.ndim == 2: # binary classification
        tp = _confusion_matrix[1][1]
        tn = _confusion_matrix[0][0]
        fp = _confusion_matrix[0][1]
        fn = _confusion_matrix[1][0]

        precision = tp / (tp + fp) if tp + fp > 0 else 0
        recall = tp / (tp + fn) if tp + fn > 0 else 0
        f_score = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

        return precision, recall, f_score
    else:
        tp_sum = _confusion_matrix[:, 1, 1]
        pred_sum = tp_sum + _confusion_matrix[:, 0, 1]
        true_sum = tp_sum + _confusion_matrix[:, 1, 0]

        # avoid nan/inf
        pred_sum[pred_sum == 0.0] = 1
        true_sum[true_sum == 0.0] = 1

        precision = tp_sum / pred_sum

        recall = tp_sum / true_sum

        f_denom = precision + recall
        f_denom[f_denom == 0.0] = 1
        f_score = 2 * (precision * recall) / f_denom

        # weighted (true labels) average
        precision = np.average(precision, weights=true_sum)
        recall = np.average(recall, weights=true_sum)
        f_score = np.average(f_score, weights=true_sum)

        return precision, recall, f_score



# coding: utf-8
from __future__ import unicode_literals

import json
import click
from pathlib import Path
from wasabi import Printer
from io import open
from .dframcy import DframCy
from .utils import get_default_columns
from .trainer import DframeTrainer, DframeEvaluator, DframeTrainClassifier

messenger = Printer()
DEFAULT_COLUMNS = ",".join(get_default_columns())


@click.group()
def main():
    pass


@main.command()
@click.option("--input_file", "-i", required=True, type=Path, help="Input text file path.")
@click.option("--output_file", "-o", required=True, type=Path, help="Output file path/name")
@click.option("--convert_type", "-t", default="csv", show_default=True, type=str, help="Output file format (JSON/CSV)")
@click.option("--language_model", "-l", default="en_core_web_sm", show_default=True, type=str, help="Language model "
                                                                                                    "to be used.")
@click.option("--columns", "-c", default=DEFAULT_COLUMNS, show_default=True, type=str, help="Annotations to be "
                                                                                            "included in dataframe.")
@click.option("--separate_entity_frame", "-s", default=False, show_default=True, type=bool, help="Save separate "
                                                                                                 "entity dataframe.")
def convert(input_file, output_file, convert_type, language_model, columns, separate_entity_frame):
    if output_file.is_dir():
        output_file = output_file.joinpath(input_file.stem + "." + str(convert_type))
    if input_file.exists():
        with open(input_file, "r") as infile:
            text = infile.read()
            dframcy = DframCy(language_model)
            doc = dframcy.nlp(text)
            if columns == DEFAULT_COLUMNS:
                annotation_dataframe = dframcy.to_dataframe(doc, separate_entity_dframe=separate_entity_frame)
            else:
                annotation_dataframe = dframcy.to_dataframe(doc, columns=columns.split(", "),
                                                            separate_entity_dframe=separate_entity_frame)
            if separate_entity_frame:
                token_annotation_dataframe, entity_dataframe = annotation_dataframe
            else:
                token_annotation_dataframe = annotation_dataframe
                entity_dataframe = None

            if convert_type == "csv":
                token_annotation_dataframe.to_csv(output_file)
                if separate_entity_frame:
                    entity_output_file = Path(str(output_file).strip(".csv") + "_entity.csv")
                    entity_dataframe.to_csv(entity_output_file)
            elif convert_type == "json":
                annotation_json = token_annotation_dataframe.to_json(orient="columns")
                with open(output_file, "w") as outfile:
                    json.dump(annotation_json, outfile)
                if separate_entity_frame:
                    entity_output_file = Path(str(output_file).strip(".json") + "_entity.json")
                    with open(entity_output_file, "w") as ent_outfile:
                        json.dump(entity_dataframe, ent_outfile)
            else:
                messenger.fail("Unknown output file format '{}'".format(convert_type), exits=-1)
    else:
        messenger.fail("input path {} does not exist".format(input_file), exits=-1)


@main.command()
@click.option("--lang", "-l", required=True, type=str, help="Model language.")
@click.option("--output_path", "-o", required=True, type=str, help="Output directory to store mode in.")
@click.option("--train_path", "-t", required=True, type=str, help="Path of CSV containing training data.")
@click.option("--dev_path", "-d", required=True, type=str, help="Path to CSV containing validation data")
@click.option("--debug_data_first", "-debug", default=True, show_default=True, type=bool, help="Run spaCy's training "
                                                                                               "data debugger before "
                                                                                               "training")
@click.option("--raw_text", "-rt", default=None, show_default=True, type=str, help="Path to jsonl file with unlabelled "
                                                                                   "text documents.")
@click.option("--base_model", "-b", default=None, show_default=True, type=str, help="Name of model to update")
@click.option("--pipeline", "-p", default="tagger,parser,ner", show_default=True, type=str, help="Comma-separated "
                                                                                                 "names of pipeline "
                                                                                                 "components")
@click.option("--vectors", "-v", default=None, show_default=True, type=str, help="Model to load vectors from")
@click.option("--n_iter", "-n", default=30, show_default=True, type=int, help="Number of iterations")
@click.option("--n_early_stopping", "-ne", default=None, show_default=True, type=int, help="Maximum number of "
                                                                                           "training epochs without "
                                                                                           "dev accuracy improvement")
@click.option("--n_examples", "-ns", default=0, show_default=True, type=int, help="Number of examples")
@click.option("--use_gpu", "-g", default=-1, show_default=True, type=int, help="Use GPU")
@click.option("--version", "-v", default="0.0.0", show_default=True, type=str, help="Model version")
@click.option("--meta_path", "-m", default=None, show_default=True, type=Path, help="Optional path to meta.json to "
                                                                                    "use as base.")
@click.option("--init_tok2vec", "-t2v", default=None, show_default=True, type=str, help="Path to pretrained weights "
                                                                                        "for the token-to-vector "
                                                                                        "parts of the models")
@click.option("--parser_multitasks", "-pm", default="", show_default=True, type=str, help="Side objectives for parser "
                                                                                          "CNN, e.g. 'dep' or 'dep,"
                                                                                          "tag'")
@click.option("--entity_multitasks", "-em", default="", show_default=True, type=str, help="Side objectives for NER "
                                                                                          "CNN, e.g. 'dep' or 'dep,"
                                                                                          "tag'")
@click.option("--noise_level", "-n", default=0.0, show_default=True, type=float, help="Amount of corruption for data "
                                                                                      "augmentation")
@click.option("--orth_variant_level", "-vl", default=0.0, show_default=True, type=float, help="Amount of orthography "
                                                                                              "variation for data "
                                                                                              "augmentation")
@click.option("--eval_beam_widths", "-bw", default="", show_default=True, type=str, help="Beam widths to evaluate, "
                                                                                         "e.g. 4,8")
@click.option("--gold_preproc", "-G", default=False, show_default=True, type=bool, help="Use gold preprocessing")
@click.option("--learn_tokens", "-T", default=False, show_default=True, type=bool, help="Make parser learn "
                                                                                        "gold-standard tokenization")
@click.option("--textcat_multilabel", "-TML", default=False, show_default=True, type=bool, help="Textcat classes "
                                                                                                "aren't mutually "
                                                                                                "exclusive ("
                                                                                                "multilabel)")
@click.option("--textcat_arch", "-ta", default="bow", show_default=True, type=str, help="Textcat model architecture")
@click.option("--textcat_positive_label", "-tpl", default=None, show_default=True, type=str, help="Textcat positive "
                                                                                                  "label for binary "
                                                                                                  "classes with two "
                                                                                                  "labels")
@click.option("--verbose", "-VV", default=False, show_default=True, type=bool, help="verbosity")
def train(
        lang,
        output_path,
        train_path,
        dev_path,
        debug_data_first,
        raw_text,
        base_model,
        pipeline,
        vectors,
        n_iter,
        n_early_stopping,
        n_examples,
        use_gpu,
        version,
        meta_path,
        init_tok2vec,
        parser_multitasks,
        entity_multitasks,
        noise_level,
        orth_variant_level,
        eval_beam_widths,
        gold_preproc,
        learn_tokens,
        textcat_multilabel,
        textcat_arch,
        textcat_positive_label,
        verbose):
    dframe_trainer = DframeTrainer(lang,
                                   output_path,
                                   train_path,
                                   dev_path,
                                   debug_data_first,
                                   raw_text,
                                   base_model,
                                   pipeline,
                                   vectors,
                                   n_iter,
                                   n_early_stopping,
                                   n_examples,
                                   use_gpu,
                                   version,
                                   meta_path,
                                   init_tok2vec,
                                   parser_multitasks,
                                   entity_multitasks,
                                   noise_level,
                                   orth_variant_level,
                                   eval_beam_widths,
                                   gold_preproc,
                                   learn_tokens,
                                   textcat_multilabel,
                                   textcat_arch,
                                   textcat_positive_label,
                                   verbose)
    dframe_trainer.begin_training()


@main.command()
@click.option("--model", "-m", required=True, type=str, help="Model name or path")
@click.option("--data_path", "-d", required=True, type=str, help="Path of CSV containing validation data")
@click.option("--gpu_id", "-g", default=-1, show_default=True, type=bool, help="Use GPU")
@click.option("--gold_preproc", "-G", default=False, show_default=True, type=bool, help="Use gold preprocessing")
@click.option("--displacy_path", "-dp", default=None, show_default=True, type=str, help="Directory to output rendered "
                                                                                        "parses as HTML")
@click.option("--displacy_limit", "-dl", default=25, show_default=True, type=int, help="Limit of parses to render as "
                                                                                       "HTML")
@click.option("--return_scores", "-R", default=False, show_default=True, type=bool, help="Return dict containing "
                                                                                         "model scores")
def evaluate(
        model,
        data_path,
        gpu_id,
        gold_preproc,
        displacy_path,
        displacy_limit,
        return_scores):
    dframe_evaluator = DframeEvaluator(
        model,
        data_path,
        gpu_id,
        gold_preproc,
        displacy_path,
        displacy_limit,
        return_scores
    )
    dframe_evaluator.begin_evaluation()


@main.command()
@click.option("--output_path", "-o", required=True, type=str, help="Output model path")
@click.option("--train_path", "-t", required=True, type=str, help="path to training data csv")
@click.option("--dev_path", "-d", required=True, type=str, help="path to testing/validation data csv")
@click.option("--model", "-m", default=None, show_default=True, type=str, help="language model name")
@click.option("--n_iter", "-n", default=20, show_default=True, type=int, help="Number of training iterations")
@click.option("--init_tok2vec", "-t2v", default=None, show_default=True, type=Path, help="Pretrained tok2vec weights")
@click.option("--exclusive_classes", "-ec", default=False, show_default=True, type=bool, help="classes exclusive")
@click.option("--architecture", "-a", default="ensemble", show_default=True, type=str, help="model architecture")
@click.option("--train_split", "-s", default=0.8, show_default=True, type=float, help="split in case no testing data")
def textcat(
        output_path,
        train_path,
        dev_path,
        model,
        n_iter,
        init_tok2vec,
        exclusive_classes,
        architecture,
        train_split):
    dframe_textcat_train = DframeTrainClassifier(
        output_path,
        train_path,
        dev_path,
        model,
        n_iter,
        init_tok2vec,
        exclusive_classes,
        architecture,
        train_split
    )
    dframe_textcat_train.begin_training()


if __name__ == '__main__':
    main()



# coding: utf-8
from __future__ import unicode_literals

import pandas as pd
from spacy.matcher import Matcher, PhraseMatcher


class DframCyMatcher(object):
    """
    Dataframe wrapper class over spaCy's Matcher
    https://spacy.io/api/matcher
    """
    def __init__(self, nlp_pipeline):
        """
        :param nlp_pipeline: nlp pipeline to be used (i.e. language model).
        """
        self._nlp = nlp_pipeline
        self._matcher = None

    @property
    def nlp(self):
        return self._nlp

    def get_nlp(self):
        return self._nlp

    def __call__(self, doc):
        """
        To find all token sequences matching the supplied patterns on the Doc
        :param doc: spacy container for linguistic annotations.
        :return: dataframe, containing matched occurrences.
        """
        df_format_json = {}
        matches = self._matcher(doc)
        for match_id, start, end in matches:
            if "match_id" not in df_format_json:
                df_format_json["match_id"] = []
                df_format_json["match_id"].append(match_id)
            else:
                df_format_json["match_id"].append(match_id)
            if "start" not in df_format_json:
                df_format_json["start"] = []
                df_format_json["start"].append(start)
            else:
                df_format_json["start"].append(start)
            if "end" not in df_format_json:
                df_format_json["end"] = []
                df_format_json["end"].append(end)
            else:
                df_format_json["end"].append(end)
            if "string_id" not in df_format_json:
                df_format_json["string_id"] = []
                df_format_json["string_id"].append(self._nlp.vocab.strings[match_id])
            else:
                df_format_json["string_id"].append(self._nlp.vocab.strings[match_id])
            if "span_text" not in df_format_json:
                df_format_json["span_text"] = []
                df_format_json["span_text"].append(doc[start:end].text)
            else:
                df_format_json["span_text"].append(doc[start:end].text)
        matches_dataframe = pd.DataFrame.from_dict(df_format_json)
        matches_dataframe.reindex(matches_dataframe["match_id"])
        matches_dataframe.drop(columns=["match_id"], inplace=True)

        return matches_dataframe

    def get_matcher_object(self):
        """
        To get spaCy's matcher class object (used for testing only).
        :return: matcher object
        """
        return self._matcher

    def get_matcher(self):
        """
        To initialize spaCy's matcher class object.
        :return: Matcher object
        """
        return Matcher(self._nlp.vocab)

    def add(self, pattern_name, callback, *pattern):
        """
        To add patterns to spaCy's matcher object
        :param pattern_name: str, pattern name
        :param callback: function, callback function to be invoked on matched occurrences.
        :param pattern: list of patterns
        """
        if not self._matcher:
            self._matcher = self.get_matcher()
        self._matcher.add(pattern_name, callback, *pattern)

    def remove(self, pattern_name):
        """
        To remove pattern from spaCy's matcher object
        :param pattern_name: str, pattern_name
        """
        if self._matcher:
            self._matcher.remove(pattern_name)

    def reset(self):
        """
        To re-initialize spaCy's matcher object
        """
        self._matcher = self.get_matcher()


class DframCyPhraseMatcher(object):
    """
        Dataframe wrapper class over spaCy's PhraseMatcher
        https://spacy.io/api/phrasematcher
    """
    def __init__(self, nlp_pipeline, attr=None):
        """
        :param nlp_pipeline: nlp pipeline to be used (i.e. language model).
        :param attr: str, token attribute to match on (default: "ORTH")
        """
        self._nlp = nlp_pipeline
        self._phrase_matcher = None
        self.attribute = attr

    @property
    def nlp(self):
        return self._nlp

    def get_nlp(self):
        return self._nlp

    def __call__(self, doc):
        """
        To find all token sequences matching the supplied patterns on the Doc
        :param doc: spacy container for linguistic annotations.
        :return: dataframe, containing matched occurrences.
        """
        df_format_json = {}
        phrase_matches = self._phrase_matcher(doc)
        for match_id, start, end in phrase_matches:
            if "match_id" not in df_format_json:
                df_format_json["match_id"] = []
                df_format_json["match_id"].append(match_id)
            else:
                df_format_json["match_id"].append(match_id)
            if "start" not in df_format_json:
                df_format_json["start"] = []
                df_format_json["start"].append(start)
            else:
                df_format_json["start"].append(start)
            if "end" not in df_format_json:
                df_format_json["end"] = []
                df_format_json["end"].append(end)
            else:
                df_format_json["end"].append(end)
            if "span_text" not in df_format_json:
                df_format_json["span_text"] = []
                df_format_json["span_text"].append(doc[start:end].text)
            else:
                df_format_json["span_text"].append(doc[start:end].text)
        phrase_matches_dataframe = pd.DataFrame.from_dict(df_format_json)
        phrase_matches_dataframe.reindex(phrase_matches_dataframe["match_id"])
        phrase_matches_dataframe.drop(columns=["match_id"], inplace=True)

        return phrase_matches_dataframe

    def get_phrase_matcher(self):
        """
        To get spaCy's phrase matcher class object (used for testing only).
        :return: phrase matcher object
        """
        return PhraseMatcher(self._nlp.vocab, attr=self.attribute) if self.attribute else PhraseMatcher(self._nlp.vocab)

    def get_phrase_matcher_object(self):
        """
        To get spaCy's matcher class object (used for testing only).
        :return: phrase matcher object
        """
        return self._phrase_matcher

    def add(self, pattern_name, callback, *pattern):
        """
        To add patterns to spaCy's phrase matcher object
        :param pattern_name: str, pattern name
        :param callback: function, callback function to be invoked on matched occurrences.
        :param pattern: list of patterns
        """
        if not self._phrase_matcher:
            self._phrase_matcher = self.get_phrase_matcher()
        self._phrase_matcher.add(pattern_name, callback, *pattern)

    def remove(self, pattern_name):
        """
        To remove pattern from spaCy's matcher object
        :param pattern_name: str, pattern_name
        """
        if self._phrase_matcher:
            self._phrase_matcher.remove(pattern_name)

    def reset(self, change_attribute=None):
        """
        To re-initialize spaCy's phrase matcher object
        :param change_attribute: token attribute to match on
        """
        self.attribute = change_attribute
        self._phrase_matcher = self.get_phrase_matcher()



# coding: utf-8
from __future__ import unicode_literals

import os
import io
import json
import spacy
import operator
import pandas as pd
import numpy as np
from wasabi import Printer
from pathlib import Path
from spacy.cli.train import train
from spacy.cli.debug_data import debug_data
from spacy.cli.evaluate import evaluate
from spacy.util import minibatch, compounding

from dframcy.dframcy import utils

messenger = Printer()


class DframeConverter(object):
    """
    Base class to convert xls/csv training data format to spaCy's JSON format
    (https://spacy.io/api/annotation#json-input).
    """

    def __init__(self,
                 train_path,
                 dev_path,
                 language_model="en_core_web_sm",
                 pipeline="tagger,parser,ner"):
        """
        initialize JSON formatter.
        :param train_path: str, file path or directory path containing multiple training data files
        :param dev_path: str, file path or directory path containing multiple validation data files
        :param language_model: str, language model to be used (default: "em_core_web_sm")
        :param pipeline: str, training pipline (default: "tagger,parser,ner")
        """
        self.train_path = train_path
        self.dev_path = dev_path
        self.dev_path = self.dev_path
        self._nlp = spacy.load(language_model)
        self.pipeline = pipeline

    def convert(self, data_path, nlp, data_type='training'):
        """
        To convert xls/csv training data to JSON format.
        :param data_path: str, single file or multiple files (directory) to be converted
        :param nlp: nlp pipeline object
        :param data_type: str, type of data "training"/"validation" (default: "training")
        :return: new JSON formatted file path, pipeline inferred from data
        """
        if os.path.exists(data_path):
            if os.path.isfile(data_path):
                if data_path.endswith(".csv"):
                    training_data = pd.read_csv(data_path)
                elif data_path.endswith(".xls") or data_path.endswith(".xlsx"):
                    excel_file = pd.ExcelFile(data_path)
                    training_data = excel_file.parse("Sheet1")
                else:
                    training_data = None
                    messenger.fail("Unknown file format for {} data file:'{}'".format(data_type, data_path), exits=-1)
            elif os.path.isdir(data_path):
                dataframe_list = []
                for file_name in os.listdir(data_path):
                    file_path = os.path.join(data_path, file_name)
                    if file_path.endswith(".csv"):
                        dataframe_list.append(pd.read_csv(file_path))
                    elif file_path.endswith(".xls") or file_path.endswith(".xlsx"):
                        excel_file = pd.ExcelFile(file_path)
                        dataframe_list.append(excel_file.parse("Sheet1"))
                    else:
                        messenger.warn("Unknown file format for {} data file:{}, skipping".format(data_type, file_path))
                training_data = pd.concat(dataframe_list, join='inner', ignore_index=True)

            training_pipeline = utils.get_training_pipeline_from_column_names(training_data.columns)
            training_pipeline = training_pipeline if len(training_pipeline.split(",")) <= len(self.pipeline.split(",")) \
                else self.pipeline

            json_format = utils.dataframe_to_spacy_training_json_format(
                training_data,
                nlp,
                training_pipeline)
            json_formatted_file_path = data_path.rstrip(".csv").rstrip(".xls").rstrip(".xlsx") + ".json"

            with io.open(json_formatted_file_path, "w") as file:
                json.dump(json_format, file)
            return json_formatted_file_path, training_pipeline
        else:
            messenger.fail("{} file/directory path does not exist".format(data_type), exits=-1)


class DframeTrainer(DframeConverter):
    """
    Wrapper class over spaCy's CLI training from CSV/XLS files.
    """

    def __init__(self,
                 lang,
                 output_path,
                 train_path,
                 dev_path,
                 debug_data_first=True,
                 raw_text=None,
                 base_model=None,
                 pipeline="tagger,parser,ner",
                 vectors=None,
                 n_iter=30,
                 n_early_stopping=None,
                 n_examples=0,
                 use_gpu=-1,
                 version="0.0.0",
                 meta_path=None,
                 init_tok2vec=None,
                 parser_multitasks="",
                 entity_multitasks="",
                 noise_level=0.0,
                 orth_variant_level=0.0,
                 eval_beam_widths="",
                 gold_preproc=False,
                 learn_tokens=False,
                 textcat_multilabel=False,
                 textcat_arch="bow",
                 textcat_positive_label=None,
                 verbose=False):
        """
        for parameters refer to: https://spacy.io/api/cli#train
        """
        super(DframeTrainer, self).__init__(train_path, dev_path, pipeline=pipeline)

        self.lang = lang
        self.output_path = output_path
        self.train_path = train_path
        self.dev_path = dev_path
        self.debug_data_first = debug_data_first
        self.raw_text = raw_text
        self.base_model = base_model
        self.pipeline = pipeline
        self.vectors = vectors
        self.n_iter = n_iter
        self.n_early_stopping = n_early_stopping
        self.n_examples = n_examples
        self.use_gpu = use_gpu
        self.version = version
        self.meta_path = meta_path
        self.init_tok2vec = init_tok2vec
        self.parser_multitasks = parser_multitasks
        self.entity_multitasks = entity_multitasks
        self.noise_level = noise_level
        self.orth_variant_level = orth_variant_level
        self.eval_beam_widths = eval_beam_widths
        self.gold_preproc = gold_preproc
        self.learn_tokens = learn_tokens
        self.textcat_multilabel = textcat_multilabel
        self.textcat_arch = textcat_arch
        self.textcat_positive_label = textcat_positive_label
        self.verbose = verbose

        if self.train_path != self.dev_path:
            self.train_path, training_pipeline = self.convert(self.train_path, self._nlp, "training")
            self.dev_path, evaluation_pipeline = self.convert(self.dev_path, self._nlp, "validation")
        else:
            messenger.warn("Same Training and validation data")
            self.train_path, training_pipeline = self.convert(self.train_path, self._nlp, "training")
            self.dev_path = self.train_path
            evaluation_pipeline = training_pipeline

        self.pipeline = training_pipeline if training_pipeline else self.pipeline
        assert training_pipeline == evaluation_pipeline, messenger.fail("Training({}) and Evaluation({}) pipeline "
                                                                        "does not "
                                                                        "match".format(training_pipeline,
                                                                                       evaluation_pipeline), exits=-1)

    def begin_training(self):
        """
        To initiate training.
        """
        if self.debug_data_first:
            debug_data(
                self.lang,
                Path(self.train_path),
                Path(self.dev_path),
                self.base_model,
                self.pipeline
            )

        train(
            self.lang,
            self.output_path,
            self.train_path,
            self.dev_path,
            self.raw_text,
            self.base_model,
            self.pipeline,
            self.vectors,
            self.n_iter,
            self.n_early_stopping,
            self.n_examples,
            self.use_gpu,
            self.version,
            self.meta_path,
            self.init_tok2vec,
            self.parser_multitasks,
            self.entity_multitasks,
            self.noise_level,
            self.orth_variant_level,
            self.eval_beam_widths,
            self.gold_preproc,
            self.learn_tokens,
            self.textcat_multilabel,
            self.textcat_arch,
            self.textcat_positive_label,
            self.verbose)


class DframeEvaluator(DframeConverter):
    """
    Wrapper class over spaCy's CLI model evaluation from CSV/XLS files.
    """

    def __init__(self,
                 model,
                 data_path,
                 gpu_id=-1,
                 gold_preproc=False,
                 displacy_path=None,
                 displacy_limit=25,
                 return_scores=False):
        """
        for parameters refer to: https://spacy.io/api/cli#evaluate
        """
        super(DframeEvaluator, self).__init__(data_path, data_path)

        self.model = model
        self.data_path = data_path
        self.gpu_id = gpu_id
        self.gold_preproc = gold_preproc
        self.displacy_path = displacy_path
        self.displacy_limit = displacy_limit
        self.return_scores = return_scores

        self.data_path, pipeline = self.convert(self.data_path, self._nlp, "evaluation")

    def begin_evaluation(self):
        """
        To initiate evaluation.
        """
        evaluate(
            self.model,
            self.data_path,
            self.gpu_id,
            self.gold_preproc,
            self.displacy_path,
            self.displacy_limit,
            self.return_scores
        )


class DframeTrainClassifier(object):
    """
    To train text classifier from CSV file.
    """

    def __init__(self,
                 output_path,
                 train_path,
                 dev_path,
                 model=None,
                 n_iter=20,
                 init_tok2vec=None,
                 exclusive_classes=False,
                 architecture="ensemble",
                 train_split=0.8):

        self.output_path = output_path
        self.train_path = train_path
        self.dev_path = dev_path
        self.model = model
        self.n_iter = n_iter
        self.init_tok2vec = init_tok2vec
        self.exclusive_classes = exclusive_classes
        self.architecture = architecture
        self.train_split = train_split
        self.label_map = None

        if self.model is not None:
            self.nlp = spacy.load(self.model)
        else:
            self.nlp = spacy.blank("en")

        if "textcat" not in self.nlp.pipe_names:
            self.textcat = self.nlp.create_pipe(
                "textcat",
                config={"exclusive_classes": self.exclusive_classes, "architecture": self.architecture}
            )
            self.nlp.add_pipe(self.textcat, last=True)
        else:
            self.textcat = self.nlp.get_pipe("textcat")

    def load_dataset(self):

        if not os.path.exists(self.train_path):
            messenger.fail("Input file path does not exist", exits=-1)

        if not os.path.exists(self.dev_path):
            messenger.warn("Validation file path does not exist, splitting training data for validation")
            self.dev_path = self.train_path

        if self.output_path is not None:
            self.output_path = Path(self.output_path)
            if not self.output_path.exists():
                self.output_path.mkdir()

        if self.train_path != self.dev_path:
            training_dataframe = pd.read_csv(self.train_path)
            testing_dataframe = pd.read_csv(self.dev_path)
        else:
            messenger.warn("Same Training and validation data")
            messenger.info("Train and test data will be split in ratio:{}".format(self.train_split))
            dataset = pd.read_csv(self.train_path)
            dataset = dataset.sample(frac=1).reset_index(drop=True)  # shuffle data
            self.train_split = int(dataset.shape[0] * self.train_split)
            datasets = np.vsplit(dataset, np.array([self.train_split]))
            training_dataframe = datasets[0]
            testing_dataframe = datasets[1]

        if not ("text" in training_dataframe.columns and "labels" in training_dataframe.columns):
            messenger.fail("Inconsistent column names in training CSV", exits=-1)

        if not ("text" in testing_dataframe.columns and "labels" in testing_dataframe.columns):
            messenger.fail("Inconsistent column names in validation CSV", exits=-1)

        unique_train_labels = list(pd.unique(training_dataframe["labels"]))
        unique_test_labels = list(pd.unique(testing_dataframe["labels"]))

        if not self.label_map:
            self.label_map = {i: l for i, l in enumerate(unique_train_labels)}

        if len(set(unique_train_labels) - set(unique_test_labels)) != 0:
            additional_labels = set(unique_train_labels) - set(unique_test_labels)
            messenger.warn("Found following additional labels in test set: {}".
                           format(", ".join(list(additional_labels))))
            messenger.info("Removing test instances having additional labels from test set")
            testing_dataframe = testing_dataframe[testing_dataframe.labels not in additional_labels]

        for label in unique_train_labels:
            self.textcat.add_label(label)

        training_text, training_label = training_dataframe["text"].tolist(), training_dataframe["labels"].tolist()
        testing_text, testing_label = testing_dataframe["text"].tolist(), testing_dataframe["labels"].tolist()

        training_cats = [{label: 1 if label == _label else 0 for label in unique_train_labels} for _label in
                         training_label]
        testing_cats = [{label: 1 if label == _label else 0 for label in unique_train_labels} for _label in
                        testing_label]
        return training_text, training_cats, testing_text, testing_cats

    def begin_training(self):

        train_text, train_label, test_text, test_label = self.load_dataset()
        train_data = list(zip(train_text, [{"cats": cats} for cats in train_label]))

        # disable other pipeline components
        other_pipes = [pipe for pipe in self.nlp.pipe_names if pipe != "textcat"]
        with self.nlp.disable_pipes(*other_pipes):
            optimizer = self.nlp.begin_training()
            if self.init_tok2vec is not None:
                with self.init_tok2vec.open("rb") as infile:
                    self.textcat.model.tok2vec.from_bytes(infile.read())
        # variable batch size, steady increase
        batch_sizes = compounding(4.0, 32.0, 1.001)

        print("{:^5}\t{:^5}\t{:^5}\t{:^5}".format("LOSS", "P", "R", "F"))

        for i in range(self.n_iter):
            losses = {}
            batches = minibatch(train_data, batch_sizes)
            for batch in batches:
                texts, annotations = zip(*batch)
                self.nlp.update(texts, annotations, sgd=optimizer, drop=0.2, losses=losses)
            with self.textcat.model.use_params(optimizer.averages):
                # evaluate on the dev data split off in load_data()
                p, r, f = self.evaluate(self.nlp.tokenizer, test_text, test_label)
                print("{0:.3f}\t{1:.3f}\t{2:.3f}\t{3:.3f}".format(losses["textcat"], p, r, f))
        with self.nlp.use_params(optimizer.averages):
            self.nlp.to_disk(self.output_path)
        messenger.info("Saved model to {}".format(self.output_path))

    def evaluate(self, tokenizer, test_texts, test_cats):
        docs = (tokenizer(text) for text in test_texts)
        true_labels = list(map(lambda x: [k for k, v in x.items() if v][0], test_cats))
        pred_labels = [max(doc.cats.items(), key=operator.itemgetter(1))[0] for doc in self.textcat.pipe(docs)]

        conf_mat = utils.confusion_matrix(pred_labels, true_labels, self.label_map)

        return utils.precision_recall_fscore(conf_mat)



# coding: utf-8
from __future__ import unicode_literals

import os
import pytest
import json
import spacy
import pandas as pd
from pandas.util.testing import assert_frame_equal
from io import open
from dframcy.dframcy import DframCy
dframcy = DframCy(spacy.load("en_core_web_sm"))

current_dir = os.path.dirname(os.path.realpath(__file__))
project_root = "/" + "/".join(current_dir.split("/")[1:-1])
data_dir = project_root + '/data'


@pytest.mark.parametrize("text", ["I am here in USA."])
def test_nlp_pipeline(text):
    doc = dframcy.nlp(text)
    assert doc[0].text == "I"
    assert doc[0].tag_ == "PRP"
    assert doc[1].lemma_ == "be"


@pytest.mark.parametrize("text", ["I am here in USA."])
def test_default_columns(text):
    doc = dframcy.nlp(text)
    dataframe = dframcy.to_dataframe(doc)
    results = pd.DataFrame({
        "tokens_start": [0, 2, 5, 10, 13, 16],
        "tokens_end": [1, 4, 9, 12, 16, 17],
        "tokens_pos": ["PRON", "AUX", "ADV", "ADP", "PROPN", "PUNCT"],
        "tokens_tag": ["PRP", "VBP", "RB", "IN", "NNP", "."],
        "tokens_dep": ["nsubj", "ROOT", "advmod", "prep", "pobj", "punct"],
        "tokens_head": [1, 1, 1, 1, 3, 1],
        "tokens_label": [None, None, None, None, "GPE", None],
        "tokens_text": ["I", "am", "here", "in", "USA", "."]
    })
    assert_frame_equal(dataframe, results)


def test_all_columns_thoroughly():
    doc = dframcy.nlp("Machine learning is an application of artificial intelligence (AI) that provides systems the "
                      "ability to automatically learn and improve from experience without being explicitly "
                      "programmed. Machine learning focuses on the development of computer programs that can access "
                      "data and use it learn for themselves.")
    dataframe = dframcy.to_dataframe(doc, ["id", "start", "end", "pos", "tag", "dep", "head", "text", "lemma", "lower",
                                           "shape", "prefix", "suffix", "is_alpha", "is_ascii", "is_digit", "is_lower",
                                           "is_upper", "is_title", "is_punct", "is_left_punct", "is_right_punct",
                                           "is_space", "is_bracket", "is_quote", "is_currency", "like_url", "like_num",
                                           "like_email", "is_oov", "is_stop", "ancestors", "conjuncts", "children",
                                           "lefts", "rights", "n_lefts", "n_rights", "is_sent_start", "has_vector",
                                           "ent_start", "ent_end", "ent_label"])
    with open(os.path.join(data_dir, "all_columns_results.json"), "r") as file:
        df_json = json.load(file)
    results = pd.DataFrame(df_json)
    assert_frame_equal(dataframe, results)


def test_entity_rule_dataframe():
    dframcy_test_ent = DframCy(spacy.load("en_core_web_sm"))
    patterns = [{"label": "ORG", "pattern": "MyCorp Inc."}]
    dframcy_test_ent.add_entity_ruler(patterns)
    doc = dframcy_test_ent.nlp("MyCorp Inc. is a company in the U.S.")
    _, entity_frame = dframcy_test_ent.to_dataframe(doc, separate_entity_dframe=True)
    results = pd.DataFrame({
        "ent_text": ["MyCorp Inc.", "U.S."],
        "ent_label": ["ORG", "GPE"]
    })
    assert_frame_equal(entity_frame, results)


def test_sentence_without_named_entities():
    doc = dframcy.nlp("Autonomous cars shift insurance liability toward manufacturers")
    dataframe = dframcy.to_dataframe(doc, ["pos", "tag", "ent_label"])

    assert "tokens_label" not in dataframe.columns



# coding: utf-8
from __future__ import unicode_literals

import io
import json
import pytest
import os
import shutil
import operator
from jsondiff import diff
from dframcy.trainer import DframeConverter, DframeTrainer, DframeEvaluator, DframeTrainClassifier


@pytest.mark.parametrize("input_csv_file, output_json_file", [("data/training_data_format.csv",
                                                               "data/training_data_converted.json")])
def test_cli_format_converter_full_pipeline(input_csv_file, output_json_file):
    dframe_converter = DframeConverter(train_path=input_csv_file, dev_path=input_csv_file)
    json_formatted_file_path, pipeline = dframe_converter.convert(input_csv_file, dframe_converter._nlp)

    with io.open(json_formatted_file_path, "r") as format_file:
        json_formatted_training_data = json.load(format_file)
    os.remove(json_formatted_file_path)

    with io.open(output_json_file, "r") as output_file:
        actual_json_formatted_training_data = json.load(output_file)

    assert diff(json_formatted_training_data, actual_json_formatted_training_data) == {}


@pytest.mark.parametrize("input_csv_file", ["data/training_data_format.csv"])
def test_cli_format_converter_only_tagger(input_csv_file):
    dframe_converter = DframeConverter(train_path=input_csv_file, dev_path=input_csv_file, pipeline="tagger")
    json_formatted_file_path, pipeline = dframe_converter.convert(input_csv_file, dframe_converter._nlp)

    training_data_only_tagger = [
        {
            "id": 0,
            "paragraphs": [
                {
                    "raw": "Uber blew through $1 million a week",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Uber",
                                    "tag": "NNP"
                                },
                                {
                                    "id": 1,
                                    "orth": "blew",
                                    "tag": "VBD"
                                },
                                {
                                    "id": 2,
                                    "orth": "through",
                                    "tag": "IN"
                                },
                                {
                                    "id": 3,
                                    "orth": "$",
                                    "tag": "$"
                                },
                                {
                                    "id": 4,
                                    "orth": "1",
                                    "tag": "CD"
                                },
                                {
                                    "id": 5,
                                    "orth": "million",
                                    "tag": "CD"
                                },
                                {
                                    "id": 6,
                                    "orth": "a",
                                    "tag": "DT"
                                },
                                {
                                    "id": 7,
                                    "orth": "week",
                                    "tag": "NN"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 1,
            "paragraphs": [
                {
                    "raw": "Android Pay expands to Canada",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Android",
                                    "tag": "NNP"
                                },
                                {
                                    "id": 1,
                                    "orth": "Pay",
                                    "tag": "NNP"
                                },
                                {
                                    "id": 2,
                                    "orth": "expands",
                                    "tag": "VBZ"
                                },
                                {
                                    "id": 3,
                                    "orth": "to",
                                    "tag": "IN"
                                },
                                {
                                    "id": 4,
                                    "orth": "Canada",
                                    "tag": "NNP"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 2,
            "paragraphs": [
                {
                    "raw": "Spotify steps up Asia expansion",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Spotify",
                                    "tag": "VB"
                                },
                                {
                                    "id": 1,
                                    "orth": "steps",
                                    "tag": "VBZ"
                                },
                                {
                                    "id": 2,
                                    "orth": "up",
                                    "tag": "RP"
                                },
                                {
                                    "id": 3,
                                    "orth": "Asia",
                                    "tag": "NNP"
                                },
                                {
                                    "id": 4,
                                    "orth": "expansion",
                                    "tag": "NN"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 3,
            "paragraphs": [
                {
                    "raw": "Google Maps launches location sharing",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Google",
                                    "tag": "NNP"
                                },
                                {
                                    "id": 1,
                                    "orth": "Maps",
                                    "tag": "NNPS"
                                },
                                {
                                    "id": 2,
                                    "orth": "launches",
                                    "tag": "VBZ"
                                },
                                {
                                    "id": 3,
                                    "orth": "location",
                                    "tag": "NN"
                                },
                                {
                                    "id": 4,
                                    "orth": "sharing",
                                    "tag": "NN"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 4,
            "paragraphs": [
                {
                    "raw": "Google rebrands its business apps",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Google",
                                    "tag": "NNP"
                                },
                                {
                                    "id": 1,
                                    "orth": "rebrands",
                                    "tag": "VBZ"
                                },
                                {
                                    "id": 2,
                                    "orth": "its",
                                    "tag": "PRP$"
                                },
                                {
                                    "id": 3,
                                    "orth": "business",
                                    "tag": "NN"
                                },
                                {
                                    "id": 4,
                                    "orth": "apps",
                                    "tag": "NNS"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 5,
            "paragraphs": [
                {
                    "raw": "look what i found on google! Joy",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "look",
                                    "tag": "VB"
                                },
                                {
                                    "id": 1,
                                    "orth": "what",
                                    "tag": "WP"
                                },
                                {
                                    "id": 2,
                                    "orth": "i",
                                    "tag": "PRP"
                                },
                                {
                                    "id": 3,
                                    "orth": "found",
                                    "tag": "VBD"
                                },
                                {
                                    "id": 4,
                                    "orth": "on",
                                    "tag": "IN"
                                },
                                {
                                    "id": 5,
                                    "orth": "google",
                                    "tag": "NNP"
                                },
                                {
                                    "id": 6,
                                    "orth": "!",
                                    "tag": "."
                                }
                            ]
                        },
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "look",
                                    "tag": "VB"
                                }
                            ]
                        }
                    ]
                }
            ]
        }
    ]

    with io.open(json_formatted_file_path, "r") as format_file:
        json_formatted_training_data = json.load(format_file)
    os.remove(json_formatted_file_path)

    assert diff(json_formatted_training_data, training_data_only_tagger) == {}


@pytest.mark.parametrize("input_csv_file", ["data/training_data_format.csv"])
def test_cli_format_converter_only_parser(input_csv_file):
    dframe_converter = DframeConverter(train_path=input_csv_file, dev_path=input_csv_file, pipeline="parser")
    json_formatted_file_path, pipeline = dframe_converter.convert(input_csv_file, dframe_converter._nlp)

    training_data_only_parser = [
        {
            "id": 0,
            "paragraphs": [
                {
                    "raw": "Uber blew through $1 million a week",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Uber",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "nsubj"
                                },
                                {
                                    "id": 1,
                                    "orth": "blew",
                                    "tag": "VBD",
                                    "head": 0,
                                    "dep": "ROOT"
                                },
                                {
                                    "id": 2,
                                    "orth": "through",
                                    "tag": "IN",
                                    "head": -1,
                                    "dep": "prt"
                                },
                                {
                                    "id": 3,
                                    "orth": "$",
                                    "tag": "$",
                                    "head": 2,
                                    "dep": "quantmod"
                                },
                                {
                                    "id": 4,
                                    "orth": "1",
                                    "tag": "CD",
                                    "head": 1,
                                    "dep": "compound"
                                },
                                {
                                    "id": 5,
                                    "orth": "million",
                                    "tag": "CD",
                                    "head": -3,
                                    "dep": "pobj"
                                },
                                {
                                    "id": 6,
                                    "orth": "a",
                                    "tag": "DT",
                                    "head": 1,
                                    "dep": "det"
                                },
                                {
                                    "id": 7,
                                    "orth": "week",
                                    "tag": "NN",
                                    "head": -2,
                                    "dep": "npadvmod"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 1,
            "paragraphs": [
                {
                    "raw": "Android Pay expands to Canada",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Android",
                                    "tag": "VB",
                                    "head": 1,
                                    "dep": "compound"
                                },
                                {
                                    "id": 1,
                                    "orth": "Pay",
                                    "tag": "NN",
                                    "head": 1,
                                    "dep": "nsubj"
                                },
                                {
                                    "id": 2,
                                    "orth": "expands",
                                    "tag": "VBZ",
                                    "head": 0,
                                    "dep": "ROOT"
                                },
                                {
                                    "id": 3,
                                    "orth": "to",
                                    "tag": "IN",
                                    "head": -1,
                                    "dep": "prep"
                                },
                                {
                                    "id": 4,
                                    "orth": "Canada",
                                    "tag": "NNP",
                                    "head": -1,
                                    "dep": "pobj"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 2,
            "paragraphs": [
                {
                    "raw": "Spotify steps up Asia expansion",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Spotify",
                                    "tag": "VB",
                                    "head": 0,
                                    "dep": "ROOT"
                                },
                                {
                                    "id": 1,
                                    "orth": "steps",
                                    "tag": "NNS",
                                    "head": -1,
                                    "dep": "dobj"
                                },
                                {
                                    "id": 2,
                                    "orth": "up",
                                    "tag": "RP",
                                    "head": -1,
                                    "dep": "prt"
                                },
                                {
                                    "id": 3,
                                    "orth": "Asia",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "compound"
                                },
                                {
                                    "id": 4,
                                    "orth": "expansion",
                                    "tag": "NN",
                                    "head": -4,
                                    "dep": "dobj"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 3,
            "paragraphs": [
                {
                    "raw": "Google Maps launches location sharing",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Google",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "compound"
                                },
                                {
                                    "id": 1,
                                    "orth": "Maps",
                                    "tag": "NNPS",
                                    "head": 1,
                                    "dep": "nsubj"
                                },
                                {
                                    "id": 2,
                                    "orth": "launches",
                                    "tag": "VBZ",
                                    "head": 0,
                                    "dep": "ROOT"
                                },
                                {
                                    "id": 3,
                                    "orth": "location",
                                    "tag": "NN",
                                    "head": 1,
                                    "dep": "compound"
                                },
                                {
                                    "id": 4,
                                    "orth": "sharing",
                                    "tag": "NN",
                                    "head": -2,
                                    "dep": "dobj"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 4,
            "paragraphs": [
                {
                    "raw": "Google rebrands its business apps",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Google",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "nsubj"
                                },
                                {
                                    "id": 1,
                                    "orth": "rebrands",
                                    "tag": "VBZ",
                                    "head": 0,
                                    "dep": "ROOT"
                                },
                                {
                                    "id": 2,
                                    "orth": "its",
                                    "tag": "PRP$",
                                    "head": 2,
                                    "dep": "poss"
                                },
                                {
                                    "id": 3,
                                    "orth": "business",
                                    "tag": "NN",
                                    "head": 1,
                                    "dep": "compound"
                                },
                                {
                                    "id": 4,
                                    "orth": "apps",
                                    "tag": "NNS",
                                    "head": -3,
                                    "dep": "dobj"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 5,
            "paragraphs": [
                {
                    "raw": "look what i found on google! Joy",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "look",
                                    "tag": "VB",
                                    "head": 0,
                                    "dep": "ROOT"
                                },
                                {
                                    "id": 1,
                                    "orth": "what",
                                    "tag": "WP",
                                    "head": 2,
                                    "dep": "dobj"
                                },
                                {
                                    "id": 2,
                                    "orth": "i",
                                    "tag": "PRP",
                                    "head": 1,
                                    "dep": "nsubj"
                                },
                                {
                                    "id": 3,
                                    "orth": "found",
                                    "tag": "VBD",
                                    "head": -3,
                                    "dep": "ccomp"
                                },
                                {
                                    "id": 4,
                                    "orth": "on",
                                    "tag": "IN",
                                    "head": -1,
                                    "dep": "prep"
                                },
                                {
                                    "id": 5,
                                    "orth": "google",
                                    "tag": "NNP",
                                    "head": -1,
                                    "dep": "pobj"
                                },
                                {
                                    "id": 6,
                                    "orth": "!",
                                    "tag": ".",
                                    "head": -6,
                                    "dep": "punct"
                                }
                            ]
                        },
                        {
                            "tokens": [
                                {
                                    "id": 7,
                                    "orth": "Joy",
                                    "tag": "NNP",
                                    "head": 0,
                                    "dep": "ROOT"
                                }
                            ]
                        }
                    ]
                }
            ]
        }
    ]

    with io.open(json_formatted_file_path, "r") as format_file:
        json_formatted_training_data = json.load(format_file)
    os.remove(json_formatted_file_path)

    assert diff(json_formatted_training_data, training_data_only_parser) == {}


@pytest.mark.parametrize("input_csv_file", ["data/training_data_format.csv"])
def test_cli_format_converter_only_ner(input_csv_file):
    dframe_converter = DframeConverter(train_path=input_csv_file, dev_path=input_csv_file, pipeline="ner")
    json_formatted_file_path, pipeline = dframe_converter.convert(input_csv_file, dframe_converter._nlp)

    training_data_only_ner = [
        {
            "id": 0,
            "paragraphs": [
                {
                    "raw": "Uber blew through $1 million a week",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Uber",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "nsubj",
                                    "ner": "U-ORG"
                                },
                                {
                                    "id": 1,
                                    "orth": "blew",
                                    "tag": "VBD",
                                    "head": 0,
                                    "dep": "ROOT",
                                    "ner": "O"
                                },
                                {
                                    "id": 2,
                                    "orth": "through",
                                    "tag": "IN",
                                    "head": -1,
                                    "dep": "prep",
                                    "ner": "O"
                                },
                                {
                                    "id": 3,
                                    "orth": "$",
                                    "tag": "$",
                                    "head": 2,
                                    "dep": "quantmod",
                                    "ner": "O"
                                },
                                {
                                    "id": 4,
                                    "orth": "1",
                                    "tag": "CD",
                                    "head": 1,
                                    "dep": "compound",
                                    "ner": "O"
                                },
                                {
                                    "id": 5,
                                    "orth": "million",
                                    "tag": "CD",
                                    "head": -3,
                                    "dep": "pobj",
                                    "ner": "O"
                                },
                                {
                                    "id": 6,
                                    "orth": "a",
                                    "tag": "DT",
                                    "head": 1,
                                    "dep": "det",
                                    "ner": "O"
                                },
                                {
                                    "id": 7,
                                    "orth": "week",
                                    "tag": "NN",
                                    "head": -2,
                                    "dep": "npadvmod",
                                    "ner": "O"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 1,
            "paragraphs": [
                {
                    "raw": "Android Pay expands to Canada",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Android",
                                    "tag": "VB",
                                    "head": 1,
                                    "dep": "compound",
                                    "ner": "B-PRODUCT"
                                },
                                {
                                    "id": 1,
                                    "orth": "Pay",
                                    "tag": "NN",
                                    "head": 1,
                                    "dep": "nsubj",
                                    "ner": "L-PRODUCT"
                                },
                                {
                                    "id": 2,
                                    "orth": "expands",
                                    "tag": "VBZ",
                                    "head": 0,
                                    "dep": "ROOT",
                                    "ner": "O"
                                },
                                {
                                    "id": 3,
                                    "orth": "to",
                                    "tag": "IN",
                                    "head": -1,
                                    "dep": "prep",
                                    "ner": "O"
                                },
                                {
                                    "id": 4,
                                    "orth": "Canada",
                                    "tag": "NNP",
                                    "head": -1,
                                    "dep": "pobj",
                                    "ner": "-"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 2,
            "paragraphs": [
                {
                    "raw": "Spotify steps up Asia expansion",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Spotify",
                                    "tag": "VB",
                                    "head": 0,
                                    "dep": "ROOT",
                                    "ner": "-"
                                },
                                {
                                    "id": 1,
                                    "orth": "steps",
                                    "tag": "NNS",
                                    "head": -1,
                                    "dep": "dobj",
                                    "ner": "O"
                                },
                                {
                                    "id": 2,
                                    "orth": "up",
                                    "tag": "RP",
                                    "head": -1,
                                    "dep": "prt",
                                    "ner": "O"
                                },
                                {
                                    "id": 3,
                                    "orth": "Asia",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "compound",
                                    "ner": "U-LOC"
                                },
                                {
                                    "id": 4,
                                    "orth": "expansion",
                                    "tag": "NN",
                                    "head": -3,
                                    "dep": "dobj",
                                    "ner": "O"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 3,
            "paragraphs": [
                {
                    "raw": "Google Maps launches location sharing",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Google",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "compound",
                                    "ner": "B-PRODUCT"
                                },
                                {
                                    "id": 1,
                                    "orth": "Maps",
                                    "tag": "NNPS",
                                    "head": 1,
                                    "dep": "nsubj",
                                    "ner": "L-PRODUCT"
                                },
                                {
                                    "id": 2,
                                    "orth": "launches",
                                    "tag": "VBZ",
                                    "head": 0,
                                    "dep": "ROOT",
                                    "ner": "O"
                                },
                                {
                                    "id": 3,
                                    "orth": "location",
                                    "tag": "NN",
                                    "head": 1,
                                    "dep": "compound",
                                    "ner": "O"
                                },
                                {
                                    "id": 4,
                                    "orth": "sharing",
                                    "tag": "NN",
                                    "head": -2,
                                    "dep": "dobj",
                                    "ner": "O"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 4,
            "paragraphs": [
                {
                    "raw": "Google rebrands its business apps",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "Google",
                                    "tag": "NNP",
                                    "head": 1,
                                    "dep": "nsubj",
                                    "ner": "U-ORG"
                                },
                                {
                                    "id": 1,
                                    "orth": "rebrands",
                                    "tag": "VBZ",
                                    "head": 0,
                                    "dep": "ROOT",
                                    "ner": "O"
                                },
                                {
                                    "id": 2,
                                    "orth": "its",
                                    "tag": "PRP$",
                                    "head": 2,
                                    "dep": "poss",
                                    "ner": "O"
                                },
                                {
                                    "id": 3,
                                    "orth": "business",
                                    "tag": "NN",
                                    "head": 1,
                                    "dep": "compound",
                                    "ner": "O"
                                },
                                {
                                    "id": 4,
                                    "orth": "apps",
                                    "tag": "NNS",
                                    "head": -3,
                                    "dep": "dobj",
                                    "ner": "O"
                                }
                            ]
                        }
                    ]
                }
            ]
        },
        {
            "id": 5,
            "paragraphs": [
                {
                    "raw": "look what i found on google! Joy",
                    "sentences": [
                        {
                            "tokens": [
                                {
                                    "id": 0,
                                    "orth": "look",
                                    "tag": "VB",
                                    "head": 0,
                                    "dep": "ROOT",
                                    "ner": "O"
                                },
                                {
                                    "id": 1,
                                    "orth": "what",
                                    "tag": "WP",
                                    "head": 2,
                                    "dep": "dobj",
                                    "ner": "O"
                                },
                                {
                                    "id": 2,
                                    "orth": "i",
                                    "tag": "PRP",
                                    "head": 1,
                                    "dep": "nsubj",
                                    "ner": "O"
                                },
                                {
                                    "id": 3,
                                    "orth": "found",
                                    "tag": "VBD",
                                    "head": -3,
                                    "dep": "ccomp",
                                    "ner": "O"
                                },
                                {
                                    "id": 4,
                                    "orth": "on",
                                    "tag": "IN",
                                    "head": -1,
                                    "dep": "prep",
                                    "ner": "O"
                                },
                                {
                                    "id": 5,
                                    "orth": "google",
                                    "tag": "NNP",
                                    "head": -1,
                                    "dep": "pobj",
                                    "ner": "U-PRODUCT"
                                },
                                {
                                    "id": 6,
                                    "orth": "!",
                                    "tag": ".",
                                    "head": -6,
                                    "dep": "punct",
                                    "ner": "O"
                                }
                            ]
                        },
                        {
                            "tokens": [
                                {
                                    "id": 7,
                                    "orth": "Joy",
                                    "tag": "NNP",
                                    "head": 0,
                                    "dep": "ROOT",
                                    "ner": "O"
                                }
                            ]
                        }
                    ]
                }
            ]
        }
    ]

    with io.open(json_formatted_file_path, "r") as format_file:
        json_formatted_training_data = json.load(format_file)
    os.remove(json_formatted_file_path)

    assert diff(json_formatted_training_data, training_data_only_ner) == {}


@pytest.mark.parametrize("input_csv_file", ["data/training_data_format.csv"])
def test_cli_training(input_csv_file):
    dframe_trainer = DframeTrainer(
        "en",
        "/tmp/",
        input_csv_file,
        input_csv_file,
        debug_data_first=False
    )
    assert dframe_trainer.pipeline == "tagger,parser,ner"


@pytest.mark.parametrize("input_csv_file", ["data/training_data_format.csv"])
def test_cli_evaluation(input_csv_file):
    dframe_evaluator = DframeEvaluator(
        "en_core_web_sm",
        input_csv_file
    )
    assert dframe_evaluator.pipeline == "tagger,parser,ner"


@pytest.mark.parametrize("input_csv_file, dev_xls_file", [("data/training_data_format.csv",
                                                          "data/training_data_format_xls.xls")])
def test_ods_training_file_format(input_csv_file, dev_xls_file):
    dframe_trainer = DframeTrainer(
        "en",
        "/tmp/",
        input_csv_file,
        dev_xls_file
    )
    os.remove(dframe_trainer.train_path)
    os.remove(dframe_trainer.dev_path)
    assert dframe_trainer.pipeline == "tagger,parser,ner"


@pytest.mark.parametrize("input_csv_file, dev_xls_file", [("data/training_data_format.csv",
                                                          "data/training_data_format_xls.xls")])
def test_training_from_directory(input_csv_file, dev_xls_file):
    if not os.path.exists("/tmp/dframcy_test"):
        os.mkdir("/tmp/dframcy_test/")

    file_name, file_extension = os.path.splitext(input_csv_file)
    file_name = str(file_name.split("/")[-1])
    for i in range(10):
        shutil.copy(input_csv_file, "/tmp/dframcy_test/" + file_name + "_" + str(i) + file_extension)

    file_name, file_extension = os.path.splitext(dev_xls_file)
    file_name = str(file_name.split("/")[-1])
    for i in range(10):
        shutil.copy(dev_xls_file, "/tmp/dframcy_test/" + file_name + "_" + str(i) + file_extension)

    dframe_trainer = DframeTrainer(
        "en",
        "/tmp/",
        "/tmp/dframcy_test/",
        "/tmp/dframcy_test/"
    )
    shutil.rmtree("/tmp/dframcy_test/")
    assert dframe_trainer.pipeline == "tagger,parser,ner"


@pytest.mark.xfail  # test case supposed to fail due to very low number of training instances
@pytest.mark.parametrize("input_csv_file", ["data/training_data_format.csv"])
def test_data_debugging(input_csv_file):
    dframe_trainer = DframeTrainer(
        "en",
        "/tmp/",
        input_csv_file,
        input_csv_file,
        debug_data_first=True
    )
    dframe_trainer.begin_training()


@pytest.mark.parametrize("input_csv_file", ["data/textcat_training.csv"])
def test_cli_textcat_training(input_csv_file):
    dframe_textcat_classifier = DframeTrainClassifier(
        "/tmp/",
        input_csv_file,
        input_csv_file,
        n_iter=1
    )
    test_text = "This movie sucked"
    dframe_textcat_classifier.begin_training()
    doc = dframe_textcat_classifier.nlp(test_text)
    assert max(doc.cats.items(), key=operator.itemgetter(1))[0] in ['NEG', 'POS']



# coding: utf-8
from __future__ import unicode_literals

import spacy
import pandas as pd
from pandas.util.testing import assert_frame_equal
from dframcy.matcher import DframCyMatcher, DframCyPhraseMatcher


dframcy_matcher = DframCyMatcher(spacy.load("en_core_web_sm"))
dframcy_phrase_matcher = DframCyPhraseMatcher(spacy.load("en_core_web_sm"), attr="LOWER")


def test_matcher():
    dframcy_matcher.reset()
    pattern = [{"LOWER": "hello"}, {"IS_PUNCT": True}, {"LOWER": "world"}]
    dframcy_matcher.add("HelloWorld", None, pattern)
    doc = dframcy_matcher.nlp(u"Hello, world! Hello world!")
    matches = dframcy_matcher.get_matcher_object()(doc)
    assert matches[0][0] == 15578876784678163569
    assert matches[0][1] == 0
    assert matches[0][2] == 3
    assert dframcy_matcher.get_nlp().vocab.strings[matches[0][0]] == "HelloWorld"
    assert doc[matches[0][1]:matches[0][2]].text == "Hello, world"

    dframcy_matcher.remove("HelloWorld")
    assert "HelloWorld" not in dframcy_matcher.get_matcher_object()


def test_matcher_dataframe():
    dframcy_matcher.reset()
    pattern = [{"LOWER": "hello"}, {"IS_PUNCT": True}, {"LOWER": "world"}]
    dframcy_matcher.add("HelloWorld", None, pattern)
    doc = dframcy_matcher.nlp(u"Hello, world! Hello world!")
    matches_dataframe = dframcy_matcher(doc)
    results = pd.DataFrame({
        "start": [0],
        "end": [3],
        "string_id": ["HelloWorld"],
        "span_text": ["Hello, world"]
    })
    assert_frame_equal(matches_dataframe, results)


def test_matcher_dataframe_multiple_patterns():
    dframcy_matcher.reset()
    dframcy_matcher.add("Hello_World", None, [{"LOWER": "hello"}, {"IS_PUNCT": True}, {"LOWER": "world"}],
                        [{"LOWER": "hello"}, {"LOWER": "world"}])
    doc = dframcy_matcher.nlp(u"Hello, world! Hello world!")
    matches_dataframe = dframcy_matcher(doc)
    results = pd.DataFrame({
        "start": [0, 4],
        "end": [3, 6],
        "string_id": ["Hello_World", "Hello_World"],
        "span_text": ["Hello, world", "Hello world"]
    })
    assert_frame_equal(matches_dataframe, results)


def test_phrase_matcher():
    patterns = [dframcy_phrase_matcher.get_nlp().make_doc(name) for name in ["Angela Merkel", "Barack Obama"]]
    dframcy_phrase_matcher.add("Names", None, *patterns)
    doc = dframcy_phrase_matcher.nlp("angela merkel and us president barack Obama")
    matches = dframcy_phrase_matcher.get_phrase_matcher_object()(doc)
    assert matches[0][0] == 10631222085860127603
    assert matches[0][1] == 0
    assert matches[0][2] == 2
    assert doc[matches[0][1]:matches[0][2]].text == "angela merkel"
    assert doc[matches[1][1]:matches[1][2]].text == "barack Obama"

    dframcy_phrase_matcher.remove("Names")
    assert "Names" not in dframcy_phrase_matcher.get_phrase_matcher_object()


def test_phrase_matcher_dataframe():
    dframcy_phrase_matcher.reset()
    terms = ["Barack Obama", "Angela Merkel", "Washington, D.C."]
    patterns = [dframcy_phrase_matcher.get_nlp().make_doc(text) for text in terms]
    dframcy_phrase_matcher.add("TerminologyList", None, *patterns)
    doc = dframcy_phrase_matcher.nlp("German Chancellor Angela Merkel and US President Barack Obama "
                                     "converse in the Oval Office inside the White House in Washington, D.C.")
    phrase_matches_dataframe = dframcy_phrase_matcher(doc)
    results = pd.DataFrame({
        "start": [2, 7, 19],
        "end": [4, 9, 22],
        "span_text": ["Angela Merkel", "Barack Obama", "Washington, D.C."]
    })
    assert_frame_equal(phrase_matches_dataframe, results)



from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    REQUIRES = [line.strip("\n") for line in f]

setup(
    name="dframcy",
    version="0.1.3",
    description="Pandas Dataframe integration for spaCy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yash1994/dframcy",
    author="Yash Patadia",
    author_email="yash@patadia.org",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
    keywords=["spacy", "dataframe", "pandas"],
    packages=find_packages(),
    install_requires=REQUIRES,
    tests_require=['pytest'],
    entry_points={
        "console_scripts": ["dframcy=dframcy.cli:main"],
    }
)



# coding: utf-8
from __future__ import unicode_literals

import pandas as pd
from spacy.pipeline import EntityRuler
from cytoolz import merge_with

from dframcy import utils


class DframCy(object):
    """
    Dataframe integration with spaCy's linguistic annotations.
    """
    def __init__(self, nlp_pipeline):
        """
        :param nlp_pipeline: nlp pipeline to be used (i.e. language model).
        """
        self._nlp = nlp_pipeline

    @property
    def nlp(self):
        return self._nlp

    @staticmethod
    def get_additional_token_attributes(attribute, doc):
        """
        To get additional (token information other than doc.json()) token class attributes.
        official doc: https://spacy.io/api/token
        type: 0 for class attribute, 1 for class method, 2 for class property
        :param attribute: tuple(name, type, is_nested) i.e. ("is_punct", 0, False)
        :param doc: spacy container for linguistic annotations.
        :return: list of or list of lists of attribute values
        """
        _name = attribute[0]
        _type = attribute[2]
        _is_nested = attribute[1]

        values = [getattr(token, _name) for token in doc]

        if _is_nested:
            if isinstance(list(values[0]), list):
                flattened_values = []
                for v in values:
                    flattened_values.append([i.text for i in v])
            else:
                flattened_values = [v.text for v in values]
            values = flattened_values
        return values

    @staticmethod
    def get_named_entity_details(doc):
        """
        To get named entities from NLP processed text
        :param doc: spacy container for linguistic annotations.
        :return: dictionary containing entity_text and entity_label
        """
        entity_details_dict = {"ent_text": [], "ent_label": []}
        for ent in doc.ents:
            entity_details_dict["ent_text"].append(ent.text)
            entity_details_dict["ent_label"].append(ent.label_)
        return entity_details_dict

    def to_dataframe(self, doc, columns=None, separate_entity_dframe=False):
        """
        Convert Linguistic annotations for text into pandas dataframe
        :param doc: spacy container for linguistic annotations.
        :param columns: list of str, name of columns to be included in dataframe (default: ["tokens.id", "tokens.text",
        "tokens.start", "tokens.end", "tokens.pos", "tokens.tag", "tokens.dep", "tokens.head", "ents.start",
        "ents.end", "ents.label"])
        :param separate_entity_dframe: bool, for separate entity dataframe (default: False)
        :return: dataframe, dataframe containing linguistic annotations
        """
        if not columns:
            columns = utils.get_default_columns()
            additional_attributes = False
        else:
            columns, additional_attributes = utils.map_user_columns_names_with_default(columns)

        if "tokens.id" not in columns:
            columns.append("tokens.id")

        json_doc = doc.to_json()

        if "ent" in ", ".join(columns):
            json_doc = utils.merge_entity_details(json_doc)

        if not json_doc["ents"]:
            columns = [cn for cn in columns if "ents" not in cn]

        merged_tokens_dict = merge_with(list, *json_doc["tokens"])
        merged_tokens_dict["text"] = [json_doc["text"][token["start"]:token["end"]] for token in json_doc["tokens"]]

        if additional_attributes:
            for column in columns:
                if column.split(".")[-1] not in merged_tokens_dict:
                    merged_tokens_dict[column.split(".")[-1]] = self.get_additional_token_attributes(
                        utils.additional_attributes_map(column), doc
                    )

        if "tokens.start" in columns and "tokens.end" in columns:
            if "ents.start" in columns:
                columns.remove("ents.start")
            if "ents.end" in columns:
                columns.remove("ents.end")

        columns_filtered_token_dict = {}

        for key in merged_tokens_dict.keys():
            if key in set([i.split(".")[-1] for i in columns]):
                columns_filtered_token_dict["tokens_" + key] = merged_tokens_dict[key]

        tokens_dataframe = pd.DataFrame.from_dict(columns_filtered_token_dict)

        tokens_dataframe.reindex(tokens_dataframe["tokens_id"])

        tokens_dataframe.drop(columns=["tokens_id"], inplace=True)

        if separate_entity_dframe:
            entity_dict = self.get_named_entity_details(doc)
            entity_dataframe = pd.DataFrame.from_dict(entity_dict)
        else:
            entity_dataframe = None

        return tokens_dataframe if not separate_entity_dframe else (tokens_dataframe, entity_dataframe)

    def add_entity_ruler(self, patterns):
        """
        To add entity ruler in nlp pipeline
        official doc: https://spacy.io/api/entityruler
        :param patterns: list or list of lists of token/phrase based patterns
        """
        ruler = EntityRuler(self._nlp)
        ruler.add_patterns(patterns)
        self._nlp.add_pipe(ruler)
